@@ -1,3 +1,5 @@
+# (c) Sound Object Logic 2000-2001
+
 use strict;
 
 package Tangram::Table;
@@ -34,31 +36,51 @@ sub new
 	my $self = bless { storage => $storage, tables => \@tables, class => $class,
 					   table_hash => $table_hash }, $pkg;
 
-	$storage->{schema}->visit_up($class,
-								 sub
-								 {
-									 my $class = shift;
+	my %seen;
+
+	for my $part ($storage->{engine}->get_parts($schema->classdef($class))) {
+	  my $table = $part->{table};
+
+	  unless (exists $seen{$table}) {
+		my $id = $seen{$table} = $storage->alloc_table;
+		#push @tables, [ $part->{name}, $id ];
+		push @tables, [ $table, $id ];
+	  }
+
+	  my $id =  $seen{$table};
+	  $table_hash->{ $part->{name} } = $id;
+
+	  $self->{root} ||= $id;
+	}
+
+	# use Data::Dumper; print Dumper \@tables;
+
+
+# 	$storage->{schema}->visit_up($class,
+# 								 sub
+# 								 {
+# 									 my $class = shift;
 			
-									 unless ($classes->{$class}{stateless})
-									 {
-										 my $id = $storage->alloc_table;
-										 push @tables, [ $class, $id ];
-										 $table_hash->{$class} = $id;
-									 }
-								 } );
+# 									 unless ($classes->{$class}{stateless})
+# 									 {
+# 										 my $id = $storage->alloc_table;
+# 										 push @tables, [ $class, $id ];
+# 										 $table_hash->{$class} = $id;
+# 									 }
+# 								 } );
 
 	return $self;
 }
 
-sub copy
-{
-	my ($pkg, $other) = @_;
+# sub copy
+# {
+# 	my ($pkg, $other) = @_;
 
-	my $self = { %$other };
-	$self->{tables} = [ @{ $self->{tables} } ];
+# 	my $self = { %$other };
+# 	$self->{tables} = [ @{ $self->{tables} } ];
 
-	bless $self, $pkg;
-}
+# 	bless $self, $pkg;
+# }
 
 sub storage
 {
@@ -71,10 +93,10 @@ sub table
 	$self->{table_hash}{$class} or confess "no table for $class in stored '$self->{class}'";
 }
 
-sub tables
-{
-	shift->{tables}
-}
+# sub tables
+# {
+# 	shift->{tables}
+# }
 
 sub class
 {
@@ -84,28 +106,33 @@ sub class
 		#return $tables->[$#$tables][0];
 }
 
-sub parts
+sub table_ids
 {
-	return map { $_->[0] } @{ shift->{tables} };
+	return map { $_->[1] } @{ shift->{tables} };
 }
 
+# sub parts
+# {
+# 	return map { $_->[0] } @{ shift->{tables} };
+# }
+
 sub root_table
 {
 	my ($self) = @_;
-	return $self->{tables}[0][1];
+	return $self->{root};
 }
 
-sub class_id_col
-{
-	my ($self) = @_;
-	return "t$self->{tables}[0][1].classId";
-}
+# sub class_id_col
+# {
+# 	my ($self) = @_;
+# 	return "t$self->{tables}[0][1].$self->{storage}{class_col}";
+# }
 
-sub leaf_table
-{
-	my ($self) = @_;
-	return $self->{tables}[-1][1];
-}
+# sub leaf_table
+# {
+# 	my ($self) = @_;
+# 	return $self->{tables}[-1][1];
+# }
 
 sub from
 {
@@ -115,7 +142,7 @@ sub from
 	my $schema = $self->storage->{schema};
 	my $classes = $schema->{classes};
 	my $tables = $self->{tables};
-	map { "$classes->{$_->[0]}{table} t$_->[1]" } @$tables;
+	map { "$_->[0] t$_->[1]" } @$tables;
 }
 
 sub where
@@ -125,60 +152,15 @@ sub where
 	my ($self) = @_;
    
 	my $tables = $self->{tables};
-	my $root = $tables->[0][1];
+	my $root = $tables->{root};
 
-	map { "t@{$_}[1].id = t$root.id" } @$tables[1..$#$tables];
+	map { "t$_->[1].id = t$root.id" } @$tables[1..$#$tables];
 }
 
-sub cols
-{
-	return join ', ', &cols unless wantarray;
-
-	my ($self) = @_;
-
-	my $tables = $self->tables;
-	my $root = $tables->[0][1];
-	my $schema = $self->storage->{schema};
-
-	my $cols = "t$root.id, t$root.classId";
-
-	foreach my $table (@$tables)
-	{
-		my ($class, $id) = @$table;
-		my $classdef = $schema->classdef($class);
-
-		foreach my $typetag (keys %{$classdef->{members}})
-		{
-			my $members = $classdef->{members}{$typetag};
-
-			foreach my $col ($schema->{types}{$typetag}->cols($members))
-			{
-				$cols .= ", t$table->[1].$col";
-			}
-		}
-	}
-
-	return $cols;
-}
-
-sub mark
-{
-	return @{ shift->{tables} };
-}
-
-sub push_spec
-{
-	my ($self, $spec) = @_;
-	my $tables = $self->tables;
-	push @$tables, [ $spec, $self->storage->alloc_table ];
-}
-
-sub pop_spec
-{
-	my ($self, $mark) = @_;
-	my $tables = $self->{tables};
-	$self->storage->free_table( map { $_->[1] } splice @$tables, $mark, @$tables - $mark );
-}
+# sub mark
+# {
+# 	return @{ shift->{tables} };
+# }
 
 sub expr_hash
 {
@@ -187,36 +169,49 @@ sub expr_hash
 	my $schema = $storage->{schema};
 	my $classes = $schema->{classes};
 	my @tables = @{$self->{tables}};
-	my $root_tid = $tables[0][1];
    
 	my %hash =
 		(
 		 object => $self, 
-		 id => Tangram::Number->expr("t$root_tid.id", $self)
+		 id => Tangram::Number->expr("t$self->{root}.$storage->{id_col}", $self),
+		 type => Tangram::Number->expr("t$self->{root}.$storage->{class_col}", $self),
 		);
 
-	$schema->visit_up($self->{class},
-					  sub
-					  {
-						  my $classdef = $classes->{shift()};
+	$hash{_IID_} = $hash{_ID_} = $hash{id};
+	$hash{_TYPE_} = $hash{type};
 
-						  my $tid = (shift @tables)->[1] unless $classdef->{stateless};
+	for my $part ($storage->{engine}->get_parts($schema->classdef($self->{class}))) {
+	  for my $field ($part->direct_fields) {
+		$hash{ $field->{name} } = $field->remote_expr($self, $self->{table_hash}{$part->{name}}, $storage);
+	  }
+	}
+													  
+	  
+	
 
-						  foreach my $typetag (keys %{$classdef->{members}})
-						  {
-							  my $type = $schema->{types}{$typetag};
-							  my $memdefs = $classdef->{members}{$typetag};
-							  @hash{$type->members($memdefs)} =
-								  $type->query_expr($self, $memdefs, $tid, $storage);
-						  }
-					  } );
+# 	$schema->visit_up($self->{class},
+# 					  sub
+# 					  {
+# 						  my $classdef = $classes->{shift()};
+
+# 						  my $tid = (shift @tables)->[1] unless $classdef->{stateless};
+
+# 						  foreach my $typetag (keys %{$classdef->{members}})
+# 						  {
+# 							  my $type = $schema->{types}{$typetag};
+# 							  my $memdefs = $classdef->{members}{$typetag};
+# 							  @hash{$type->members($memdefs)} =
+# 								  $type->query_expr($self, $memdefs, $tid, $storage);
+# 						  }
+# 					  } );
 
 	return \%hash;
 }
 
 package Tangram::RDBObject;
 
-use base qw( Tangram::CursorObject );
+use vars qw(@ISA);
+ @ISA = qw( Tangram::CursorObject );
 
 sub where
 {
@@ -242,7 +237,7 @@ sub where
 		$schema->for_each_spec($class,
 							   sub { my $spec = shift; push @class_ids, $storage->class_id($spec) unless $classes->{$spec}{abstract} } );
 
-		@where_class_id = "t$root.classId IN (" . join(', ', $storage->_kind_class_ids($class) ) . ')';
+		@where_class_id = "t$root.$storage->{class_col} IN (" . join(', ', $storage->_kind_class_ids($class) ) . ')';
 	}
 
 	return (@where_class_id, map { "t@{$_}[1].id = t$root.id" } @$tables[1..$#$tables]);
@@ -294,7 +289,102 @@ sub as_string
 	return ref($self) . "($self->{expr})";
 }
 
-use overload "&" => \&and, "|" => \&or, '!' => \&not, fallback => 1;
+
+sub sum
+{
+  my ($self, $val) = @_;
+
+  # $DB::single = 1;
+
+  Tangram::Expr->new(Tangram::Number->instance,
+		     "SUM(" . $self->{expr} . ")",
+		     $self->objects,
+		     );
+
+}
+
+
+# BEGIN ks.perl@kurtstephens.com 2002/06/25
+sub unaop
+{
+    Tangram::Expr::unaop(@_);
+}
+
+
+sub binop
+{
+    my ($self, $op, $arg, $tight, $swap) = @_;
+
+    my @objects = $self->objects;
+    my $objects = Set::Object->new(@objects);
+    # my $storage = $self->{storage};
+    my $ltight = $self->{'tight'};
+    my $rtight = 100;
+
+    if ( ref($arg) ) {
+	if ( $arg->isa('Tangram::Expr') ) {
+	    $objects->insert($arg->objects);
+	    $rtight = $arg->{'tight'};
+	    $arg = $arg->{'expr'};
+	}
+	if ( $arg->isa('Tangram::Filter') ) {
+	    $objects->insert($arg->objects);
+	    $rtight = $arg->{'tight'};
+	    $arg = $arg->{'expr'};
+	}
+	elsif ( $arg->isa('Tangram::QueryObject') ) {
+	    $objects->insert($arg->object);
+	    $rtight = $arg->{'tight'};
+	    $arg = $arg->{'id'}->{'expr'};
+	}
+    }
+
+    $tight ||= 100;
+    $self = $self->{'expr'};
+    $self = "($self)" if $ltight < $tight;
+    $arg  = "($arg)"  if $rtight < $tight;
+    if ( $swap ) {
+      ($self, $arg) = ($arg, $self);
+    }
+    # $DB::single = $swap;
+
+    return new Tangram::Filter(expr => "$self $op $arg", tight => $tight,
+			       objects => $objects );
+}
+
+
+# Aliases
+*cos =  \&Tangram::Expr::sin;
+*sin =  \&Tangram::Expr::cos;
+*acos = \&Tangram::Expr::acos;
+
+#use overload "&" => \&and, "|" => \&or, '!' => \&not, fallback => 1;
+use overload 
+  "&"    => \&and, 
+  "|"    => \&or, 
+  '!'    => \&not,
+  '+'    => \&Tangram::Expr::add,
+  '-'    => \&Tangram::Expr::subt,
+  '*'    => \&Tangram::Expr::mul,
+  '/'    => \&Tangram::Expr::div,
+  'cos'  => \&Tangram::Expr::cos, 
+  'sin'  => \&Tangram::Expr::sin,
+  'acos' => \&Tangram::Expr::acos,
+  "=="   => \&Tangram::Expr::eq,
+  "eq"   => \&Tangram::Expr::eq,
+  "!="   => \&Tangram::Expr::ne,
+  "ne"   => \&Tangram::Expr::ne,
+  "<"    => \&Tangram::Expr::lt,
+  "lt"   => \&Tangram::Expr::lt,
+  "<="   => \&Tangram::Expr::le,
+  "le"   => \&Tangram::Expr::le,
+  ">"    => \&Tangram::Expr::gt,
+  "gt"   => \&Tangram::Expr::gt,
+  ">="   => \&Tangram::Expr::ge,
+  "ge"   => \&Tangram::Expr::ge,
+  fallback => 1;
+# END ks.perl@kurtstephens.com 2002/06/25
+
 
 sub op
 {
@@ -340,6 +430,7 @@ sub objects
 }
 
 package Tangram::Expr;
+use Carp;
 
 sub new
 {
@@ -356,7 +447,7 @@ sub expr
 
 sub storage
 {
-	return shift->{objects}[0]->{storage};
+	return ((shift->{objects}->members)[0] or confess 'no storage')->storage;
 }
 
 sub type
@@ -381,39 +472,101 @@ sub ne
 	return $self->binop('<>', $arg);
 }
 
+# BEGIN ks.perl@kurtstephens.com 2002/06/25
 sub lt
 {
-	my ($self, $arg) = @_;
-	return $self->binop('<', $arg);
+	my ($self, $arg, $swap) = @_;
+	return $self->binop('<', $arg, undef, $swap);
 }
 
 sub le
 {
-	my ($self, $arg) = @_;
-	return $self->binop('<=', $arg);
+	my ($self, $arg, $swap) = @_;
+	return $self->binop('<=', $arg, undef, $swap);
 }
 
 sub gt
 {
-	my ($self, $arg) = @_;
-	return $self->binop('>', $arg);
+	my ($self, $arg, $swap) = @_;
+	return $self->binop('>', $arg, undef, $swap);
 }
 
 sub ge
 {
-	my ($self, $arg) = @_;
-	return $self->binop('>=', $arg);
+	my ($self, $arg, $swap) = @_;
+	return $self->binop('>=', $arg, undef, $swap);
 }
 
+sub add
+{
+    my ($self, $arg) = @_;
+    $self->binop('+', $arg, 90);
+}
+
+
+sub subt
+{
+    my ($self, $arg, $swap) = @_;
+    $self->binop('-', $arg, 90, $swap);
+}
+
+
+sub mul
+{
+    my ($self, $arg) = @_;
+    $self->binop('*', $arg, 95);
+}
+
+
+sub div
+{
+    my ($self, $arg, $swap) = @_;
+    $self->binop('/', $arg, 95, $swap);
+}
+
+
+sub cos
+{
+    my ($self) = @_;
+    $self->unaop('COS', 100);
+}
+
+
+sub sin
+{
+    my ($self) = @_;
+    $self->unaop('SIN', 100);
+}
+
+sub acos
+{
+    my ($self) = @_;
+    $self->unaop('ACOS', 100);
+}
+
+
+sub unaop
+{
+    my ($self, $op, $tight) = @_;
+    
+    my @objects = $self->objects;
+    my $objects = Set::Object->new(@objects);
+    my $storage = $self->{storage};
+    
+    return new Tangram::Filter(expr => "$op($self->{expr})", tight => $tight || 100,
+			       objects => $objects );
+}
+
+
 sub binop
 {
-	my ($self, $op, $arg) = @_;
+	my ($self, $op, $arg, $tight, $swap) = @_;
 
 	my @objects = $self->objects;
 	my $objects = Set::Object->new(@objects);
 	my $storage = $self->{storage};
 
-	if ($arg)
+	if (defined $arg)
 	{
 		if (my $type = ref($arg))
 		{
@@ -431,7 +584,7 @@ sub binop
    
 			elsif (exists $storage->{schema}{classes}{$type})
 			{
-				$arg = $storage->id($arg) or Carp::confess "$arg is not persistent";
+				$arg = $storage->export_object($arg) or Carp::confess "$arg is not persistent";
 			}
 
 			else
@@ -450,9 +603,14 @@ sub binop
 		$arg = 'NULL';
 	}
 
-	return new Tangram::Filter(expr => "$self->{expr} $op $arg", tight => 100,
+	my ($l, $r) = $swap ? ($arg, $self->{expr}) : ($self->{expr}, $arg);
+	$tight ||= 100;
+
+	return new Tangram::Filter(expr => "$l $op $r", tight => $tight,
 							   objects => $objects );
 }
+# END ks.perl@kurtstephens.com 2002/06/25
+
 
 sub like
 {
@@ -464,7 +622,7 @@ sub like
 sub count
 {
 	my ($self, $val) = @_;
-	$self->{storage}{dialect}
+	$self->{storage}
 		->expr(Tangram::Integer->instance, "COUNT($self->{expr})",
 				$self->objects );
 }
@@ -475,7 +633,29 @@ sub as_string
 	return ref($self) . "($self->{expr})";
 }
 
+sub DESTROY { }
+
+use vars qw( $AUTOLOAD );
+
+sub AUTOLOAD {
+  my $fun = $AUTOLOAD;
+  $fun =~ s/.*:://;
+  
+  my $self = shift;
+  my $expr = $self->expr(); # the SQL string for this Expr
+  $self->{type}->expr("$fun($expr)", $self->objects);
+}
+
 use overload
+# BEGIN ks.perl@kurtstephens.com 2002/06/25
+        '+'    => \&add,
+        '-'    => \&subt,
+        '*'    => \&mul,
+        '/'    => \&div,
+        'cos'  => \&cos, 
+        'sin'  => \&sin,
+        'acos' => \&acos,
+# END ks.perl@kurtstephens.com 2002/06/25
 	"==" => \&eq,
 	"eq" => \&eq,
 	"!=" => \&ne,
@@ -506,6 +686,11 @@ sub object
 	shift->{object}
 }
 
+sub table_ids
+{
+	shift->{object}->table_ids()
+}
+
 sub class
 {
 	shift->{object}{class}
@@ -527,7 +712,7 @@ sub eq
 	{
 		my $other_id = $self->{object}{storage}->id($other)
 			or confess "'$other' is not a persistent object";
-		$self->{id} == $self->{object}{storage}->id($other)
+		$self->{id} == $self->{object}{storage}->export_object($other)
 	}
 }
 
@@ -537,20 +722,43 @@ sub is_kind_of
 
 	my $object = $self->{object};
 	my $root = $object->{tables}[0][1];
+	my $storage = $object->{storage};
 
 	Tangram::Filter->new(
-						 expr => "t$root.classId IN (" . join(', ', $object->{storage}->_kind_class_ids($class) ) . ')',
+						 expr => "t$root.$storage->{class_col} IN (" . join(', ', $storage->_kind_class_ids($class) ) . ')',
 						 tight => 100,
 						 objects => Set::Object->new( $object ) );
 }
 
+
+sub expr
+{
+  shift->{id}{expr}
+}
+
+
+sub count
+{
+  my ($self, $val) = @_;
+
+  # $DB::single = 1;
+
+  Tangram::Expr->new(Tangram::Integer->instance,
+		     "COUNT(" . $self->{id}{expr} . ")",
+		     $self->{id}->objects,
+		     );
+
+}
+
+
 use overload "==" => \&eq, "!=" => \&ne, fallback => 1;
 
 package Tangram::Select;
 
 use Carp;
 
-use base qw( Tangram::Expr );
+use vars qw(@ISA);
+ @ISA = qw( Tangram::Expr );
 
 sub new
 {
@@ -585,6 +793,11 @@ sub new
 	my $sql = "SELECT";
 	$sql .= ' DISTINCT' if $args{distinct};
 	$sql .= "  $cols";
+	if (exists $args{order}) {
+	    $sql .= join("", map {", $_"}
+			 grep { $sql !~ m/ \Q$_\E(?:,|$)/ }
+			 map { $_->{expr} } @{$args{order}});
+	}
 	$sql .= "\nFROM $from" if $from;
 	$sql .= "\nWHERE $where" if $where;
 
